package at.bestsolution.cef;

import java.net.URL;
import java.util.concurrent.CompletableFuture;

public interface Cef {
	
	/**
     * Initialize CEF.
     * 
     * Only use this method directly when running the message loop manually, i.e. via {@link #doMessageLoopWork()}.
     * Must be called from the main application thread (see CEF documentation on <code>CefInitialize</code>).
     */
    public void initialize();

    /**
     * Start CEF and run the automatic message loop.
     * 
     * A new thread will be started which initializes CEF and then runs the 
     * CEF message loop (see <code>CefRunMessageLoop</code>). When using this method,
     * {@link #initialize()} must not be invoked manually as it will be handled by the
     * implementation to ensure it happens on the correct thread.
     * 
     * @return a {@link CompletableFuture} which completes when CEF has been initialized.
     * Can be used to determine the status as alternative to polling {@link #isInitialized()}.
     */
    public CompletableFuture<Void> start();

    /**
     * Perform a single iteration of CEF message loop processing.
     * 
     * Can be used to integrate the CEF message loop into an existing message loop. CEF must be initialized manually
     * via {@link #initialize()} beforehand. Must be called from the main application thread 
     * (see CEF documentation on <code>CefDoMessageLoopWork</code>).
     * 
     * Note that care needs to be taken to neither call this too infrequently (to avoid starving the CEF message loop)
     * but also not too often (to avoid impacting CPU usage) - see CEF documentation for details when using this approach.
     */
    public void doMessageLoopWork();

    /**
     * Shutdown CEF processes, should be called before the application exits.
     * 
     * When operating using a manual message loop via {@link #doMessageLoopWork()}, this must be called from the main
     * application thread (see CEF documentation on <code>CefShutdown</code>). When using the integrated/automatic
     * message loop it may be invoked from any thread as the implementation ensures the shutdown is run on the correct thread.
     * 
     * @throws InterruptedException if the current thread is interrupted whilst waiting on the 
     *                              shutdown of the integrated message loop
     */
    public void shutdown() throws InterruptedException;

    public boolean isInitialized();

    /**
     * Create a new browser.
     * 
     * @param renderHandler handler which provides viewport information and renders images
     * @param initialURL the initial URL to load in the created browser
     * @return a new browser instance
     * 
     * @throws IllegalStateException if CEF has not been initialized yet
     */
    public Browser createBrowser(RenderHandler renderHandler, URL initialURL);

}
