package at.bestsolution.cef.javafx;

import java.lang.reflect.Method;
import java.net.URL;
import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicReference;

import at.bestsolution.cef.Browser;
import at.bestsolution.cef.Cef;
import at.bestsolution.cef.Rectangle;
import at.bestsolution.cef.RenderHandler;
import at.bestsolution.cef.ScreenInfo;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ObservableList;
import javafx.geometry.Bounds;
import javafx.geometry.Rectangle2D;
import javafx.scene.image.ImageView;
import javafx.scene.image.PixelFormat;
import javafx.scene.image.WritableImage;
import javafx.scene.layout.Region;
import javafx.stage.Screen;
import javafx.stage.Window;

public class BrowserView extends Region {

    private final Browser browser;
    private final ImageView renderingView;

    private final AtomicReference<Rectangle> viewRect;

    public BrowserView(Cef cef, URL initialURL) {
        renderingView = new ImageView();
        renderingView.setPreserveRatio(false);
        getChildren().add(renderingView);

        viewRect = new AtomicReference<>(new Rectangle(0, 0, 10, 10));
        browser = cef.createBrowser(new RenderHanderImpl(), initialURL);

        setOnMouseMoved(evt -> browser.sendMouseMoveEvent((int) evt.getX(), (int) evt.getY(), false));
        setOnMouseExited(evt -> browser.sendMouseMoveEvent((int) evt.getX(), (int) evt.getY(), true));
        setOnScroll(evt -> browser.sendMouseWheelEvent((int) evt.getX(), (int) evt.getY(), (int) evt.getDeltaX(), (int) evt.getDeltaY()));
        setOnMousePressed(evt -> browser.sendMouseClickEvent((int) evt.getX(), (int) evt.getY(), false, 1));
        setOnMouseDragged(evt -> browser.sendMouseMoveEvent((int) evt.getX(), (int) evt.getY(), false));
        setOnMouseReleased(evt -> browser.sendMouseClickEvent((int) evt.getX(), (int) evt.getY(), true, 1));

        this.layoutBoundsProperty().addListener(new ChangeListener<Bounds>() {
            @Override
            public void changed(ObservableValue<? extends Bounds> observable, Bounds oldValue, Bounds newValue) {
                viewRect.set(new Rectangle(0, 0, Math.max((int) newValue.getWidth(), 5), Math.max((int) newValue.getHeight(), 5)));
                browser.wasResized();
            }
        });
    }

    public void close() {
        browser.close();
    }

    public void reloadPage(boolean ignoreCache) {
        browser.reload(ignoreCache);
    }

    public void loadPage(URL url) {
        browser.load(url);
    }

    private class RenderHanderImpl implements RenderHandler {

        @Override
        public void onPaint(ByteBuffer buffer, int width, int height) {
            WritableImage image = new WritableImage(width, height);
            image.getPixelWriter().setPixels(0, 0, width, height, PixelFormat.getByteBgraInstance(), buffer, width * 4);
            ScreenInfo screenInfo = getScreenInfo();
            float scale = screenInfo != null ? screenInfo.deviceScaleFactor : 1;
            Platform.runLater(() -> {
            	renderingView.setImage(image);
            	renderingView.setFitWidth(width/ scale);
            	renderingView.setFitHeight(height/ scale);
            });
        }

        @Override
        public Rectangle getViewRect() {
            return viewRect.get();
        }

        @Override
        public ScreenInfo getScreenInfo() {
            return Utils.platformRunSync(() -> {
                Window window = getScene().getWindow();
                ObservableList<Screen> screens = Screen.getScreensForRectangle(window.getX(), window.getY(), window.getWidth(), window.getHeight());
                if (screens.size() < 1) {
                    return null;
                }
                // select the first found screen - may need a better method of detection in case multiple screen use different scaling / DPI values,
                // however can only pass one setting to Cef and would need to merge retrieved values
                Screen currentScreen = screens.get(0);

                float scaleFactor = getScreenScaleFactor(currentScreen);
                Rectangle rect = toRectangle(currentScreen.getBounds());
                Rectangle availableRect = toRectangle(currentScreen.getVisualBounds());

                // TODO determine how to retrieve thes color depth settings from the system
                int colorDepth = 32;
                int colorDepthPerComponent = 8;

                return new ScreenInfo(scaleFactor, colorDepth, colorDepthPerComponent, false, rect, availableRect);
            }, () -> null );
        }
    }

    private static Rectangle toRectangle(Rectangle2D rect) {
        return new Rectangle((int) rect.getMinX(), (int) rect.getMinY(), (int) rect.getWidth(), (int) rect.getHeight());
    }

    private static float getScreenScaleFactor(Screen screen) {
        try {
            // JavaFX >= 9, access via reflection to avoid errors when running on older versions
            Method m = Screen.class.getMethod("getOutputScaleX");
            return ((Double) m.invoke(screen)).floatValue();
        } catch (ReflectiveOperationException | IllegalArgumentException e) {
            try {
                // JavaFX 8 - internal API: ScreenHelper.getScreenAccessor().getRenderScale(screen)
                Class<?> screenHelperClass = Class.forName("com.sun.javafx.stage.ScreenHelper");
                Method getScreenAccessor = screenHelperClass.getMethod("getScreenAccessor");
                Object screenAccessorInstance = getScreenAccessor.invoke(null);

                Class<?> screenAccessorClass = Class.forName("com.sun.javafx.stage.ScreenHelper$ScreenAccessor");
                Method getRenderScale = screenAccessorClass.getMethod("getRenderScale", Screen.class);
                return (Float) getRenderScale.invoke(screenAccessorInstance, screen);
            } catch (ReflectiveOperationException | IllegalArgumentException eInner) {
                return 1.0f;
            }
        }
    }
}
