package at.bestsolution.cef.javafx;

import java.util.concurrent.CompletableFuture;

import at.bestsolution.cef.Cef;
import at.bestsolution.cef.util.CefMessageLoopManager;
import at.bestsolution.cef.util.CefMessageLoopManagerFactory;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Platform;
import javafx.util.Duration;

/**
 * Ensures CEF message loop (and the initialization of CEF) run on the JavaFX App Thread.
 * This is required for MacOS - events passed from OS such as theme events otherwise 
 * end up on the wrong thread, as they must arrive via the "main" thread in CEF but 
 * always originate from the FX thread on MacOS
 */
public class JfxAppThreadMessageLoopManagerFactory implements CefMessageLoopManagerFactory {

    public static final int DEFAULT_MESSAGE_LOOP_FRAME_RATE = 30;

    private final int msgLoopFrameRate;

    public JfxAppThreadMessageLoopManagerFactory() {
        this(DEFAULT_MESSAGE_LOOP_FRAME_RATE);
    }

    public JfxAppThreadMessageLoopManagerFactory(int msgLoopFrameRate) {
        this.msgLoopFrameRate = msgLoopFrameRate;
    }

    @Override
    public CefMessageLoopManager createMessageLoopManager(Cef cef) {
        final Timeline msgLoopTimeline = new Timeline(new KeyFrame(
            Duration.seconds(1).divide(msgLoopFrameRate),
            "CEF Message Loop",
            e -> cef.doMessageLoopWork() ));
        msgLoopTimeline.setCycleCount(Animation.INDEFINITE);

        return new CefMessageLoopManager() {

            private final Timeline messageLoop = msgLoopTimeline;

            @Override
            public CompletableFuture<Void> startMessageLoop() {
            	CompletableFuture<Void> initFuture = new CompletableFuture<>();
            	Platform.runLater(() -> {
            		cef.initialize();
            		initFuture.complete( null );
            		messageLoop.play();
            	});
            	
            	return initFuture;
            }

            @Override
            public void stopMessageLoop() throws InterruptedException {
                messageLoop.stop();
                cef.shutdown();
            }
            
        };
    }
    
}
