package at.bestsolution.cef.javafx;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import java.util.concurrent.RunnableFuture;
import java.util.function.Function;
import java.util.function.Supplier;

import javafx.application.Platform;

class Utils {
	
	private Utils() {}
	
	/**
	 * Perform a computation on the JavaFX Application Thread and return its result. Blocks until the result is available.
	 * 
	 * @param <U> return type
	 * @param computation the computation to perform on the Application Thread
	 * @param defaultValueSupplier supplies a default value which will be returned upon failure of the computation
	 * @return the result produced by the computation
	 */
	public static <U> U platformRunSync( Supplier<U> computation, Supplier<U> defaultValueSupplier ) {
		return platformRunSync( ignored -> computation.get(), null, defaultValueSupplier );
	}
	
	/**
	 * Perform a computation on the JavaFX Application Thread using some input and return its result.
	 * Blocks until the result is available.
	 * 
	 * @param <T> type of the input
	 * @param <U> return type
	 * @param computation the computation to perform on the Application Thread
	 * @param input input for the computation operation
	 * @param defaultValueSupplier supplies a default value which will be returned upon failure of the computation
	 * @return the result produced by the computation
	 */
	public static <T,U> U platformRunSync( Function<T,U> computation, T input, Supplier<U> defaultValueSupplier ) {
		RunnableFuture<U> future = new FutureTask<>( () -> computation.apply( input ) );
    	Platform.runLater( future );
    	
    	try {
			return future.get();
    	} catch( InterruptedException e ) {
            Thread.currentThread().interrupt();
            return defaultValueSupplier.get();
    	} catch( ExecutionException e ) {
    		return defaultValueSupplier.get();
    	}
	}
	
}
