package at.bestsolution.cef.jni;

import java.io.IOException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.PosixFilePermissions;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.bestsolution.cef.Browser;
import at.bestsolution.cef.Cef;
import at.bestsolution.cef.RenderHandler;
import at.bestsolution.cef.util.OperatingSystem;
import at.bestsolution.cef.util.ResourceHelper;

public class JNICef implements Cef {

    private static final Logger LOGGER = LoggerFactory.getLogger(JNICef.class);

    private static final CountDownLatch CEF_NATIVE_LOADED;
    private static String nativeExtractionDir;

    private Thread mainThread;
    private final AtomicBoolean isInitialized;

    private static Path unpackAndLoadNative() throws IOException {
        Path libDir = ResourceHelper.getExtractedNativeResources();

        // TODO better library loading - maybe look at how glass does it?
        switch( OperatingSystem.getType() ) {
        	case WINDOWS:
				List<String> winLibs = Arrays.asList("d3dcompiler_47", "libEGL", "libGLESv2", "swiftshader/libEGL", "swiftshader/libGLESv2", "chrome_elf", "libcef", "cef-fx");
	            for (String libName : winLibs) {
	                Path lib = libDir.resolve(libName + ".dll");
	                LOGGER.debug("Using {}", lib);
	                System.load(lib.toString());
	            }
	            break;
        	case MACOS:
        		List<String> macLibs = Arrays.asList("cef-fx.native-shared", "cef-fx.native-jni");
                for (String libName : macLibs) {
                    Path lib = libDir.resolve("lib" + libName + ".dylib");
                    LOGGER.debug("Using {}", lib);
                    System.load(lib.toString());
                }
                Arrays.asList("", " (GPU)", " (Plugin)", " (Renderer)").forEach( helper -> {
                    Path subprocessBin = libDir.resolve(String.format("CEF-fx.app/Contents/Frameworks/CEF-fx Helper%s.app/Contents/MacOS/CEF-fx Helper%s", helper, helper));
                    try {
                        Files.setPosixFilePermissions(subprocessBin, PosixFilePermissions.fromString("rwxr-x---"));
                    } catch (IOException e) {
                        throw new IllegalStateException(e);
                    }
                });
        		break;
        	case LINUX:
        		List<String> linuxLibs = Arrays.asList("libEGL", "libGLESv2", "swiftshader/libEGL", "swiftshader/libGLESv2", "libcef", "libcef-fx.native-shared", "libcef-fx.native-jni");
                for (String libName : linuxLibs) {
                    Path lib = libDir.resolve(libName + ".so");
                    LOGGER.debug("Using {}", lib);
                    System.load(lib.toString());
                }
                Arrays.asList("cef-fx.native-helper", "helper.sh").forEach( helper -> {
                    Path subprocessBin = libDir.resolve(helper);
                    try {
                        Files.setPosixFilePermissions(subprocessBin, PosixFilePermissions.fromString("rwxr-x---"));
                    } catch (IOException e) {
                        throw new IllegalStateException(e);
                    }
                });
        		break;
			default:
				throw new IllegalStateException("unsupported platform");
		}

        return libDir;
    }

    static {
        CEF_NATIVE_LOADED = new CountDownLatch(1);

        // unpacking native code may take a while, run in background - access to extraction dir
        // protected by latch
        CompletableFuture.runAsync(() -> {
            try {
                nativeExtractionDir = unpackAndLoadNative().toAbsolutePath().toString();
            } catch (IOException e) {
                throw new CompletionException(e);
            }
        }).whenComplete((result, e) -> {
            if (e != null) {
                LOGGER.error("failed to unpack and load native files", e);
            }
            CEF_NATIVE_LOADED.countDown();
        });
        
    }

    public JNICef() {
        mainThread = null;
        isInitialized = new AtomicBoolean( false );
    }

    @Override
    public void initialize() {
        try {
            // wait until native code extraction is complete
            CEF_NATIVE_LOADED.await();

            LOGGER.info("Native CEF files located at {}", nativeExtractionDir);
            if (!initializeNative(nativeExtractionDir)) {
                throw new RuntimeException("CEF initialization failed");
            }
            isInitialized.set( true );
        } catch(InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new RuntimeException(e);
        }
    }

    private native boolean initializeNative(String cefExtractionDir);

    @Override
    public CompletableFuture<Void> start() {
        if(isInitialized.get()) {
            throw new IllegalStateException("CEF already initialized");
        }
        
        CompletableFuture<Void> initializationFuture = new CompletableFuture<>();
        mainThread = new Thread(() -> {
            // initialize, message loop and shutdown should be called from the main application thread
            // as per CEF docs
        	try {
        		initialize();
        		initializationFuture.complete( null );
        	} catch (RuntimeException e) {
        		initializationFuture.completeExceptionally( e );
        		return;
        	}
            runMessageLoop(); // blocks until quitMessageLoop is invoked
            shutdownNative();
        }, "CEF-main");
        mainThread.start();
        
        return initializationFuture;
    }

    private native void runMessageLoop();
    private native void quitMessageLoop();

    @Override
    public native void doMessageLoopWork();

    @Override
    public void shutdown() throws InterruptedException {
        if(mainThread != null) {
            // using the CEF builtin message loop -> initiate shutdown and wait for completion,
            // actual shutdown and cleanup handled within the thread before it exits
            quitMessageLoop();
            mainThread.join();
            mainThread = null;
        } else {
            // using custom message loop -> shutdown CEF directly
            shutdownNative();
        }
        isInitialized.set( false );
    }
    private native void shutdownNative();

    public boolean isInitialized() {
        return isInitialized.get();
    }

    protected void ensureInitialized() {
        if (!isInitialized.get()) {
            throw new IllegalStateException("CEF not initialized");
        }
    }

    @Override
    public Browser createBrowser(RenderHandler renderHandler, URL initialURL) {
        ensureInitialized();
        return new JNIBrowser(renderHandler, initialURL);
    }
}
