package at.bestsolution.cef.process;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.ProcessBuilder.Redirect;
import java.nio.file.Path;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.UnaryOperator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.bestsolution.cef.util.OperatingSystem;

class ExternalProcessManager extends Thread {
	
	private static final Logger LOGGER = LoggerFactory.getLogger(ExternalProcessManager.class);

	private static final Pattern COMMAND_PATTERN = Pattern.compile(":(\\w+)\\|(\\w+)\\|(.*)");
	private static final Pattern RESPONSE_PATTERN = Pattern.compile(">(!)?(\\w+)\\|(.*)");
	
	private final Process process;
	private volatile boolean exit;

	private BufferedWriter output;
	private final ReentrantLock outputLock = new ReentrantLock();
	
	private final AtomicInteger nextCommandId = new AtomicInteger(0);
	private final AtomicInteger nextEntityId = new AtomicInteger(0);
	private final ConcurrentHashMap<String, CompletableFuture<String>> pendingCommands = new ConcurrentHashMap<>();
	
	private final ConcurrentHashMap<String, UnaryOperator<String>> commandHandlers = new ConcurrentHashMap<>();

	private static Path getExternalProcessPath( Path cefExtractionDir ) {
		switch (OperatingSystem.getType()) {
			case LINUX:
				return cefExtractionDir.resolve( "standalone.sh" );
			case WINDOWS:
				return cefExtractionDir.resolve( "cef-fx-standalone.exe" );
			default:
				return cefExtractionDir.resolve( "cef-fx.native-standalone" );
		}
		
	}
	
	public ExternalProcessManager( Path cefExtractionDir ) throws IOException {
		ProcessBuilder processBuilder = new ProcessBuilder(
				getExternalProcessPath( cefExtractionDir ).toAbsolutePath().toString(),
				cefExtractionDir.toAbsolutePath().toString() )
			.directory( cefExtractionDir.toFile() )
			.redirectOutput( Redirect.PIPE )
			.redirectInput( Redirect.PIPE )
			.redirectError( Redirect.INHERIT );
		
		exit = false;
		output = null;
		
		process = processBuilder.start();
		output = new BufferedWriter( new OutputStreamWriter( process.getOutputStream() ) );
	}
	
	@Override
	public void run() {
		try {
			try ( BufferedReader input = new BufferedReader( new InputStreamReader( process.getInputStream() ) ); ) {
				while (!exit && process.isAlive()) {
					String line = input.readLine();
					if (line == null) {
						// input stream ended - process exited
					} else {
						Matcher commandMatcher = COMMAND_PATTERN.matcher( line );
						Matcher responseMatcher = RESPONSE_PATTERN.matcher( line );
						if (commandMatcher.matches()) {
							handleCommand( commandMatcher.group( 1 ), commandMatcher.group( 2 ), commandMatcher.group( 3 ) );
						} else if (responseMatcher.matches()) {
							handleResponse( responseMatcher.group( 2 ), responseMatcher.group( 1 ) != null, responseMatcher.group( 3 ) );
						} else {
							LOGGER.error( "received unexpected data from subprocess: {}", line );
						}
					}
				}
			}
			process.waitFor();
		} catch( IOException e ) {
			LOGGER.error("IO error during communication with external process", e);
		} catch( InterruptedException e ) {
			interrupt();
			LOGGER.info("external process manager interrupted", e);
		}
	}
	
	public void shutdown() {
		exit = true;
		sendCommand( "exit", null );
		interrupt();
	}
	
	public void registerCommandHandler(String command, UnaryOperator<String> handler) {
		commandHandlers.put( command, handler );
	}
	
	public void removeCommandHandler(String command) {
		commandHandlers.remove( command );
	}
	
	public CompletableFuture<String> sendCommand(String command, String payload) {
		String id = Integer.toHexString( nextCommandId.getAndIncrement() );
		CompletableFuture<String> responseFuture = new CompletableFuture<>();
		pendingCommands.put( id, responseFuture );
		
		outputLock.lock();
		try {
			output.write( ":" );
			output.write( id );
			output.write( "|" );
			output.write( command );
			output.write( "|" );
			output.write( payload != null ? payload : "" );
			output.newLine();
			output.flush();
		} catch( IOException e ) {
			LOGGER.error("error whilst sending command to external process", e);
		} finally {
			outputLock.unlock();
		}
		
		return responseFuture;
	}

	/**
	 * Generate a new entity ID, e.g. to setup commands for a specific instance of something (e.g. a specific "browser").
	 * Ensures IDs used for entities are unique within the context of this process manager and thus avoids
	 * conflicts when routing commands to handlers.
	 * @return a new ID which is unique for this manager.
	 */
	public String getNextEntityId() {
		return Integer.toHexString( nextEntityId.getAndIncrement() );
	}
	
	private void sendResponse(String id, boolean isError, String payload) {
		outputLock.lock();
			try {
				output.write( ">" );
				if (isError) {
					output.write("!");
				}
				output.write( id );
				output.write( "|" );
				output.write( payload != null ? payload : "" );
				output.newLine();
				output.flush();
			} catch( IOException e ) {
				LOGGER.error("error whilst sending response to external process", e);
			} finally {
				outputLock.unlock();
			}
	}

	private void handleCommand(String id, String command, String payload) {
		UnaryOperator<String> handler = commandHandlers.get( command );
		if (handler != null) {
			String response = handler.apply( payload );
			sendResponse(id, false, response);
		} else {
			LOGGER.warn( "received unknown command with id '{}' (no handler registered): '{}'", id, command );
			sendResponse( id, true, "no handler registered for command" );
		}
	}
	
	private void handleResponse(String id, boolean isError, String payload) {
		CompletableFuture<String> responseFuture = pendingCommands.remove( id );
		if (responseFuture != null) {
			if (!isError) {
				responseFuture.complete( payload );
			} else {
				responseFuture.completeExceptionally( new RuntimeException( payload ) );
			}
		} else {
			LOGGER.error( "received response with unknown ID {}", id );
		}
	}
}
