package at.bestsolution.cef.process;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.file.StandardOpenOption;
import java.util.Locale;
import java.util.concurrent.ExecutionException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.bestsolution.cef.Browser;
import at.bestsolution.cef.Point;
import at.bestsolution.cef.Rectangle;
import at.bestsolution.cef.RenderHandler;
import at.bestsolution.cef.ScreenInfo;
import at.bestsolution.cef.util.OperatingSystem;

class ProcessBrowser implements Browser {

	private static final Logger LOGGER = LoggerFactory.getLogger(ProcessBrowser.class);

	private static final long SHARED_BUFFER_SIZE = 1024L * 1024 * 50;
	
	private final ExternalProcessManager processManager;
    private final String nativeBrowserHandle;
    
    private final File sharedBufferFile;
    private final MappedByteBuffer sharedBuffer;
    
    public ProcessBrowser(ExternalProcessManager processManager, RenderHandler renderHandler, URL initialURL) throws InterruptedException, ExecutionException, IOException {
    	this.processManager = processManager;
		this.nativeBrowserHandle = processManager.getNextEntityId();
    	
    	File sharedBufferFileDirectory = OperatingSystem.isLinux() ? new File( "/dev/shm" ) : null;
    	sharedBufferFile = File.createTempFile( "beso-cef_browser_", ".buffer", sharedBufferFileDirectory );
    	sharedBufferFile.deleteOnExit();
    	try( FileChannel fc = FileChannel.open( sharedBufferFile.toPath(), StandardOpenOption.READ, StandardOpenOption.WRITE, StandardOpenOption.CREATE )) {
    		sharedBuffer = fc.map( FileChannel.MapMode.READ_ONLY, 0, SHARED_BUFFER_SIZE);
    	}		
    	
		registerRenderHandlerCommands( processManager, renderHandler );
		processManager.sendCommand( "browsercreate",
					nativeBrowserHandle
					+ "|" + sharedBufferFile.getAbsolutePath()
					+ "|" + Long.toString( SHARED_BUFFER_SIZE )
					+ "|" + ( initialURL != null ? initialURL.toString() : "about:blank" ) )
			.get();
    }
    
    private void registerRenderHandlerCommands(ExternalProcessManager processManager, RenderHandler renderHandler) {
    	processManager.registerCommandHandler( "browser_getRootScreenRect_" + nativeBrowserHandle, params -> rectangleToPayload( renderHandler.getRootScreenRect() ) );
    	
    	processManager.registerCommandHandler( "browser_getScreenInfo_" + nativeBrowserHandle, params -> {
    		ScreenInfo info = renderHandler.getScreenInfo();
    		if (info == null) {
    			return "null";
    		} else {
    			return String.format( Locale.ROOT, // format parameters include a float -> ensure decimal separator char is not localized for compatibility with native code
    					"%s,%s,%s,%s,%s,%s",
    					info.deviceScaleFactor,
    					info.colorDepth,
    					info.colorDepthPerComponent,
    					booleanToPayload( info.isMonochrome ),
    					rectangleToPayload( info.rect ),
    					rectangleToPayload( info.availableRect )
    				);
    		}
    	});
    	
    	processManager.registerCommandHandler( "browser_getScreenPoint_" + nativeBrowserHandle, params -> {
    		int delimiterPos = params.indexOf( "," );
    		int x = Integer.parseInt( params.substring( 0, delimiterPos ) );
    		int y = Integer.parseInt( params.substring( delimiterPos + 1, params.length() ) );
    		
    		Point point = renderHandler.getScreenPoint( x, y );
    		if (point == null) {
    			return "null";
    		} else {
    			return String.format( "%s,%s", point.x, point.y );
    		}
    	});
    	
    	processManager.registerCommandHandler( "browser_getViewRect_" + nativeBrowserHandle, params -> rectangleToPayload( renderHandler.getViewRect() ) );
    	
    	processManager.registerCommandHandler( "browser_onPaint_" + nativeBrowserHandle, params -> {
    		int delimiterPos = params.indexOf( "," );
    		int width = Integer.parseInt( params.substring( 0, delimiterPos ) );
    		int height = Integer.parseInt( params.substring( delimiterPos + 1, params.length() ) );
    		
    		sharedBuffer.rewind();
    		renderHandler.onPaint( sharedBuffer, width, height );
    		
    		return null;
    	});
    }

    @Override
    public void close() {
        processManager.sendCommand( "browserclose", nativeBrowserHandle ).join();
        
        processManager.removeCommandHandler( "browser_getRootScreenRect_" + nativeBrowserHandle );
        processManager.removeCommandHandler( "browser_getScreenInfo_" + nativeBrowserHandle );
        processManager.removeCommandHandler( "browser_getScreenPoint_" + nativeBrowserHandle );
        processManager.removeCommandHandler( "browser_getViewRect_" + nativeBrowserHandle );
        processManager.removeCommandHandler( "browser_onPaint_" + nativeBrowserHandle );
    }

    @Override
    public boolean isReady() {
		try {
			return Boolean.parseBoolean( processManager.sendCommand( "browser_ready_" + nativeBrowserHandle, null ).get() );
		} catch( InterruptedException | ExecutionException e ) {
			LOGGER.warn("failed to check browser readiness", e);
			return false;
		}
	}

    @Override
    public void wasResized() {
    	processManager.sendCommand( "browser_evt-resized_" + nativeBrowserHandle, null );
    }
    
    @Override
    public void sendMouseMoveEvent(int x, int y, boolean mouseLeave) {
    	processManager.sendCommand( "browser_evt-mousemove_" + nativeBrowserHandle, String.format( "%s,%s,%s", x, y, booleanToPayload( mouseLeave ) ) );
    }
    
    @Override
    public void sendMouseWheelEvent(int x, int y, int deltaX, int deltaY) {
    	processManager.sendCommand( "browser_evt-mousewheel_" + nativeBrowserHandle, String.format( "%s,%s,%s,%s", x, y, deltaX, deltaY ) );
    }
    
    @Override
    public void sendMouseClickEvent(int x, int y, boolean mouseUp, int clickCount) {
    	processManager.sendCommand( "browser_evt-mouseclick_" + nativeBrowserHandle, String.format( "%s,%s,%s,%s", x, y, booleanToPayload( mouseUp ), clickCount ) );
    }
    
	@Override
	public void reload(boolean ignoreCache) {
		processManager.sendCommand( "browser_reload_" + nativeBrowserHandle, booleanToPayload( ignoreCache ) ).join();
	}

	@Override
	public void load(URL url) {
		processManager.sendCommand( "browser_load-url_" + nativeBrowserHandle, url != null ? url.toString() : "about:blank" ).join();
	}

    private static String booleanToPayload( boolean value ) {
    	return value ? "t" : "f";
    }
    
    private static String rectangleToPayload( Rectangle rect ) {
    	if (rect == null) {
    		return "null";
    	} else {
    		return String.format( "%s,%s,%s,%s", rect.x, rect.y, rect.width, rect.height );
    	}
    }
}
