package at.bestsolution.cef.process;

import java.io.IOException;
import java.net.URL;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicBoolean;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.bestsolution.cef.Browser;
import at.bestsolution.cef.Cef;
import at.bestsolution.cef.RenderHandler;
import at.bestsolution.cef.util.OperatingSystem;
import at.bestsolution.cef.util.ResourceHelper;

public class ProcessCef implements Cef {

    private static final Logger LOGGER = LoggerFactory.getLogger(ProcessCef.class);

    private static final CountDownLatch CEF_NATIVE_LOADED;
    private static Path nativeExtractionDir;

    private ExternalProcessManager processManager;
    private AtomicBoolean messageLoopRunning = new AtomicBoolean( false );
    
    private static Path unpackNative() throws IOException {
        Path libDir = ResourceHelper.getExtractedNativeResources();

        switch( OperatingSystem.getType() ) {
			case WINDOWS:
				// no further actions required on Windows
				break;
			case MACOS:
				ResourceHelper.ensureExecutable( libDir.resolve("cef-fx.native-standalone") );
	            Arrays.asList("", " (GPU)", " (Plugin)", " (Renderer)").forEach( helper -> {
	                ResourceHelper.ensureExecutable( libDir.resolve(String.format("CEF-fx.app/Contents/Frameworks/CEF-fx Helper%s.app/Contents/MacOS/CEF-fx Helper%s", helper, helper)) );
	            });
				break;
			case LINUX:
				Arrays.asList("cef-fx.native-standalone", "standalone.sh", "cef-fx.native-helper", "helper.sh").forEach( helper -> {
	                ResourceHelper.ensureExecutable( libDir.resolve(helper) );
	            });
				break;
			default:
				throw new IllegalStateException("unsupported platform");
		}

        return libDir;
    }

	static {
        CEF_NATIVE_LOADED = new CountDownLatch(1);

        // unpacking native code may take a while, run in background - access to extraction dir
        // protected by latch
        CompletableFuture.runAsync(() -> {
            try {
                nativeExtractionDir = unpackNative();
            } catch (IOException e) {
                throw new CompletionException(e);
            }
        }).whenComplete((result, e) -> {
            if (e != null) {
                LOGGER.error("failed to unpack native files", e);
            }
            CEF_NATIVE_LOADED.countDown();
        });
        
    }

    public ProcessCef() {
    	processManager = null;
    }

    @Override
    public void initialize() {
    	try {
	        // wait until native code extraction is complete
	        CEF_NATIVE_LOADED.await();
	        LOGGER.info("Native CEF files located at {}", nativeExtractionDir);
	        
	        processManager = new ExternalProcessManager( nativeExtractionDir );
	        processManager.start();
    	} catch(IOException e) {
    		throw new RuntimeException(e);
    	} catch(InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new RuntimeException(e);
        }
    }

    @Override
    public CompletableFuture<Void> start() {
        if (isInitialized()) {
            throw new IllegalStateException("CEF already initialized");
        }
        
        initialize();
        return processManager.sendCommand( "cefstartmsgloop", null )
        		.thenRun ( () -> messageLoopRunning.set( true ) );
    }
    
    @Override
    public void doMessageLoopWork() {
    	ensureInitialized();
    	processManager.sendCommand( "cefdomsgloopwork", null ).join();
    }

    @Override
    public void shutdown() throws InterruptedException {
    	if (messageLoopRunning.get()) {
    		processManager.sendCommand( "cefquitmsgloop", null ).join();
    	}
    	processManager.shutdown();
    	processManager.join();
    }

    public boolean isInitialized() {
    	return processManager != null && processManager.isAlive();
    }

    protected void ensureInitialized() {
        if (!isInitialized()) {
        	throw new IllegalStateException("CEF not initialized or external process not running");
        }
    }

    @Override
    public Browser createBrowser(RenderHandler renderHandler, URL initialURL) {
        ensureInitialized();
        try {
			return new ProcessBrowser( processManager, renderHandler, initialURL );
		} catch ( InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new RuntimeException(e);
		} catch ( ExecutionException | IOException e ) {
			throw new RuntimeException(e);
		}
    }
}
