package at.bestsolution.cef.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.PosixFilePermissions;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.bestsolution.cef.Cef;

public class ResourceHelper {

    private static final Logger LOGGER = LoggerFactory.getLogger(ResourceHelper.class);

	private ResourceHelper() {}
	
	protected static String readResourceHash(Class<?> resourceClass, String baseDir) throws IOException {
        URL hashResource = resourceClass.getResource(baseDir + "meta/contenthash.txt");
        try(BufferedReader reader = new BufferedReader(new InputStreamReader(hashResource.openStream(), StandardCharsets.UTF_8))) {
            return reader.readLine();
        }
    }

    public static Path getExtractedNativeResources() throws IOException {
        final String preextractedPath = System.getProperty("cef-fx-native-dir");
        if( preextractedPath != null  ) {
			Path libDir = Paths.get(System.getProperty("cef-fx-native-dir"));
			if( Files.exists(libDir) ) {
				return libDir;
			} else {
                LOGGER.warn("CEF native files configured to be loaded from '{}', but directory is empty. Will try to temporarily extract the files instead.", preextractedPath);
            }
		}

        return unpackResourceList( Cef.class, "native/" );
    }
	
	public static Path unpackResourceList(Class<?> resourceClass, String baseDir) throws IOException {
        String hash = readResourceHash(resourceClass, baseDir);

        Path tmpDir = Paths.get(System.getProperty("java.io.tmpdir"));
        Path libDir = tmpDir.resolve("beso-cef-native_" + hash);

        if (Files.notExists(libDir)) {
            URL fileListResource = resourceClass.getResource(baseDir + "meta/filelist.txt");
            if (fileListResource == null) {
                throw new IOException("file index list not found in specified resource directory");
            }

            try(BufferedReader reader = new BufferedReader(new InputStreamReader(fileListResource.openStream(), StandardCharsets.UTF_8))) {
                String filePath;
                while ((filePath = reader.readLine()) != null) {
                    URL fileResource = resourceClass.getResource(baseDir + filePath);
                    if (fileResource == null) {
                        LOGGER.warn("file '{}' specified in index list not found in resources, skipping");
                    } else {
                        Path file = libDir.resolve(filePath);
                        if (!file.toAbsolutePath().startsWith( libDir.toAbsolutePath() )) {
                            // prevent writing to arbitrary locations in case the used file index list is corrupted or contains malicious/modified content
                            throw new IllegalStateException("file entry outside target dir: " + filePath);
                        }

                        LOGGER.debug("Extracting {}", file);
                        Files.createDirectories(file.getParent());
                        try(InputStream fileInput = fileResource.openStream()) {
                            Files.copy(fileInput, file);
                        }
                    }
                }
            }
        }
        
        return libDir;
	}

	public static void ensureExecutable( Path path ) {
		try {
		    Files.setPosixFilePermissions(path, PosixFilePermissions.fromString("rwxr-x---"));
		} catch (IOException e) {
		    throw new IllegalStateException(e);
		}
	}
	
}
